package port

import (
	"net/http"

	"gitea.dwysokinski.me/twhelp/corev3/internal/domain"
	"gitea.dwysokinski.me/twhelp/corev3/internal/port/internal/apimodel"
)

func (h *apiHTTPHandler) ListTribes(
	w http.ResponseWriter,
	r *http.Request,
	_ apimodel.VersionCodePathParam,
	serverKey apimodel.ServerKeyPathParam,
	params apimodel.ListTribesParams,
) {
	domainParams := domain.NewListTribesParams()

	if err := domainParams.SetSort([]domain.TribeSort{domain.TribeSortIDASC}); err != nil {
		apiErrorRenderer{errors: []error{err}, formatErrorPath: formatListTribesParamsErrorPath}.render(w, r)
		return
	}

	if err := domainParams.SetServerKeys([]string{serverKey}); err != nil {
		apiErrorRenderer{errors: []error{err}, formatErrorPath: formatListTribesParamsErrorPath}.render(w, r)
		return
	}

	if params.Deleted != nil {
		if err := domainParams.SetDeleted(domain.NullBool{
			Value: *params.Deleted,
			Valid: true,
		}); err != nil {
			apiErrorRenderer{errors: []error{err}, formatErrorPath: formatListTribesParamsErrorPath}.render(w, r)
			return
		}
	}

	if params.Limit != nil {
		if err := domainParams.SetLimit(*params.Limit); err != nil {
			apiErrorRenderer{errors: []error{err}, formatErrorPath: formatListTribesParamsErrorPath}.render(w, r)
			return
		}
	}

	if params.Cursor != nil {
		if err := domainParams.SetEncodedCursor(*params.Cursor); err != nil {
			apiErrorRenderer{errors: []error{err}, formatErrorPath: formatListTribesParamsErrorPath}.render(w, r)
			return
		}
	}

	res, err := h.tribeSvc.List(r.Context(), domainParams)
	if err != nil {
		apiErrorRenderer{errors: []error{err}}.render(w, r)
		return
	}

	renderJSON(w, r, http.StatusOK, apimodel.NewListTribesResponse(res))
}

func formatListTribesParamsErrorPath(segments []errorPathSegment) []string {
	if segments[0].model != "ListTribesParams" {
		return nil
	}

	switch segments[0].field {
	case "cursor":
		return []string{"$query", "cursor"}
	case "limit":
		return []string{"$query", "limit"}
	case "deleted":
		return []string{"$query", "deleted"}
	case "sort":
		return []string{"$query", "sort"}
	default:
		return nil
	}
}
